/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "hardware/adc.h"
#include "hardware/gpio.h"
#include "hardware/watchdog.h"
#include "pico/bootrom.h"
#include "pico/stdlib.h"
#include "pico/time.h"

#include "device.h"
#include "meter.h"

// GPIO
static const uint GPIO_PIN[] = {
    BTN1, BTN2, BTN3, BTN4, BTN5, BTN6, BTN7, BTN8, BTN9,    // ボタン
    TA_RE1, TA_RE2,                                          // ロータリーエンコーダ
};

// 電圧値
static float   pcV = 0.0f;           // PC電圧
static int16_t samplingCount = 0;    // SAMPLING_uS間の電圧サンプリング数

// TA
static float pcTA = TA_MAX;           // PCのTA値
static float setTA = TA_MIN;          // 基準TA値
static float taList[TA_HIST_SIZE];    // PCのTAの履歴
static float taRange = 0.0f;          // 履歴内のTA変動範囲

// TAカウンター
static float  taCounter = 0.0f;                // TAカウンター
static float  counterList[COUNT_HIST_SIZE];    // TAカウンター履歴
static size_t counterListIndex = 0;            // TAカウンター履歴のインデックス

// 感度
static float sens[2] = {0.0f, 0.0f};    // 感度 [0]:通常モード、[1]:低感度モード
typedef enum {
    SENS_NORM = 0,
    SENS_LOW = 1
} sensIndex_e;
static sensIndex_e sensIndex = SENS_NORM;    // 感度配列のインデックス
static float       sensGain = 1.0f;          // TA値による感度の補正倍率

// メーター
static float meterValue = 0.0f;      // メーターの目盛値  左端:0、セット位置:15、右端:40
static float taScaleRatio = 0.0f;    // 1TAあたりのメーター目盛変化量
static float dacPos = 0;             // DACの値

// 自動セット用
static int64_t autoSetWaitUs = AUTO_SET_WAIT_uS;    // 待ち時間

// イベントフラグ
static bool setEvt = false;       // セット実行
static bool sensEvt = false;      // 感度操作
static bool updateEvt = false;    // LCD表示更新

// 状態管理
static bool autoSet = false;         // 自動セット有効
static bool counting = false;        // TAカウント有効
static bool pause = false;           // TAカウントを中断
static bool powerView = false;       // 電源表示モード
static bool standby = true;          // 節電中
static bool usbConnected = false;    // USB接続フラグ
static bool timeExceeded = false;    // タイマー用
static bool outOfRange = false;      // 針がSAMPLING_uS間連続して範囲外ならtrue

/**
 * @brief 電圧からTAを計算する。
 */
static inline float v2ta(float v) {
    return ((v / CONST_V) * (TA_MAX - TA_MIN)) + TA_MIN;
}

/**
 * @brief TAから抵抗値を計算する。
 */
static inline float ta2r(float ta) {
    return CONST_R * (ta - TA_MIN) / (TA_MAX - ta);
}

/**
 * @brief 抵抗値からTAを計算する。
 */
static inline float r2ta(float r) {
    return r / (r + CONST_R) * (TA_MAX - TA_MIN) + TA_MIN;
}

/**
 * @brief 引数のTAのときの感度補正係数
 * 
 * TA2.0、感度6.4(補正前)でFALL目盛までリードする抵抗値(450Ω)を基準抵抗値とする。
 * 引数のTA値にてPCの抵抗値が基準抵抗値下がった際のTAの変化量を返す。
 * 
 */
static inline float rTaRatio(float ta) {
    float subR = ta2r(ta) - 450.0f;
    float subTA = r2ta(subR);
    return ta - subTA;
}

/**
 * @brief 現在の感度および基準TA値における1TAあたりの補正されたメーター目盛の変化量を計算する。
 *
 * 感度の定義：
 *   ・5kΩ(TA2.0)の抵抗を接続し針をセット位置に設定した状態で、100kΩの抵抗を
 * 　　並列接続した際に、針がメモリの右端まで移動する増幅率が感度32となる。
 *   ・同様に50kΩを並列接続した際に、針がメモリの右端まで移動する増幅率が感度16となる。
 * ⇒ 感度Sのとき、[S × 3.125kΩ]の抵抗を並列接続すると、針がメモリの左端を指す。
 *
 * 感度が補正されていない場合、基準TA値によらずTAの変化量と針の振れ幅は比例する。
 */
static void calcScaleRatio() {
    // 現在の感度における1TAあたりのメーター目盛の変化量を計算
    float r = 1.0f / (1.0f / (sens[sensIndex] * 3125.0f) + 1.0f / 5000.0f);    // 針を右端まで振れさせる抵抗値
    float dTA = 2.0f - r2ta(r);                                                // 針を右端まで振れさせるTA変化量
    float dScale = ((METER_SCALE_SIZE - METER_SCALE_SET) / dTA);               // 補正前の1TAあたりの目盛変化量

    // 現在の基準TAにおける補正倍率を計算
    float tTA = setTA;
    // 基準TAがCORRECT_TA_LOW_THからCORRECT_TA_HIGH_THを外れる場合は補正を穏やかにする
    if (tTA < CORRECT_TA_LOW_TH) {
        tTA = CORRECT_TA_LOW_TH - (CORRECT_TA_LOW_TH - tTA) * CORRECT_TA_LOW_RATE;
    } else if (tTA > CORRECT_TA_HIGH_TH) {
        tTA = CORRECT_TA_HIGH_TH + (tTA - CORRECT_TA_HIGH_TH) * CORRECT_TA_HIGH_RATE;
    }
    sensGain = rTaRatio(CORRECT_TA_BASE) / rTaRatio(tTA);    // CORRECT_TA_BASEに対する現在の基準TAでの補正倍率を計算する。
    sensGain = sensGain / CORRECT_COEF_TA2;                  // 感度の定義はTA2.0を基準とするため、TA2.0の補正を1倍とする。

    taScaleRatio = dScale * sensGain;    // 補正後の1TAあたりメーター目盛の変化量
}

/**
 * @brief TAアクションをカウントする。
 */
static void countTA() {
    if (!pause && counting) {
        float dTA = (setTA - pcTA);
        if (dTA > 0) {
            taCounter += dTA;
            ++counterListIndex;
            counterListIndex = counterListIndex % COUNT_HIST_SIZE;
            counterList[counterListIndex] = taCounter;
        }
    }
}

/**
 * @brief TAカウントを履歴から復元する
 */
static void undoCounter(int num) {
    counterListIndex = (counterListIndex + COUNT_HIST_SIZE + num) % COUNT_HIST_SIZE;
    taCounter = counterList[(counterListIndex + COUNT_HIST_SIZE) % COUNT_HIST_SIZE];
}

/**
 * @brief PCのTA履歴を保存する。履歴からTAの変動範囲を集計する。
 */
static void addTaList(float ta) {
    static size_t listIndex = 0;

    taList[listIndex] = ta;
    float maxValue = 0.0f;
    float minValue = 7.0f;
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        if (maxValue < taList[i]) {
            maxValue = taList[i];
        }
        if (minValue > taList[i]) {
            minValue = taList[i];
        }
    }
    taRange = fabs(maxValue - minValue);

    listIndex = (listIndex + 1) % TA_HIST_SIZE;
}

/**
 * @brief PCのTA履歴ををクリアする。
 *
 */
static void clearTaHist() {
    taRange = 0.0f;
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        taList[i] = pcTA;
    }
}

/**
 * @brief アラームハンドラ
 */
static int64_t alarmHandler(alarm_id_t id, void *user_data) {
    timeExceeded = true;
    return 0;
}

/**
 * @brief ボタンイベントを管理し実行可否を判定する。
 *
 * 　前回の実行後 EVENT_WAIT_uS 以上経過している場合trueを返す
 */
static bool checkBtnAction(uint target, uint gpio, uint32_t event) {
    static uint64_t nextAction[29];    // gpio:0-28
    bool            result = false;
    if (target == gpio) {
        uint64_t nowUs = time_us_64();
        result = ((nowUs > nextAction[target]) && (event == GPIO_IRQ_EDGE_FALL));
        nextAction[target] = nowUs + EVENT_WAIT_uS;
    }
    return result;
}

/**
 * @brief GPIO割り込みハンドラ
 */
static void irqHandler(uint gpio, uint32_t events) {
    // ロータリーエンコーダー
    if (gpio == TA_RE1 || gpio == TA_RE2) {
        static const int8_t RE_DIR[] = ROTARY_DIRECTION;
        static uint8_t      taRE1;
        static uint8_t      taRE2;
        static size_t       reIndex = 0;

        uint8_t state = (events == GPIO_IRQ_EDGE_RISE) ? 1 : 0;
        taRE1 = (gpio == TA_RE1) ? state : taRE1;
        taRE2 = (gpio == TA_RE2) ? state : taRE2;

        size_t lastIndex = reIndex;
        reIndex = ((reIndex << 2) + (taRE1 << 1) + taRE2) & 0xF;
        if (lastIndex != reIndex) {
            int8_t delta = RE_DIR[reIndex];
            if (delta != 0) {
                if ((delta > 0) && (sens[sensIndex] < SENS_MAX)) {
                    sens[sensIndex] = sens[sensIndex] * SENS_COEF;
                    sensEvt = true;
                } else if ((delta < 0) && (SENS_MIN < sens[sensIndex])) {
                    sens[sensIndex] = sens[sensIndex] / SENS_COEF;
                    sensEvt = true;
                }
            }
        }
    }

    // セットボタン
    else if (checkBtnAction(BUTTON_SET, gpio, events)) {
        setEvt = true;
        pause = false;
    }

    // セットボタンその２
    else if (checkBtnAction(BUTTON_SET2, gpio, events)) {
        setEvt = true;
        pause = false;
    }

    // 電源表示ボタン
    else if (gpio == BUTTON_POWER) {
        updateEvt = true;
    }

    // 低感度ボタン
    else if (checkBtnAction(BUTTON_SENS0, gpio, events)) {
        // sensIndex = (sensIndex + 1) & 1;
        if (sensIndex == SENS_NORM) {
            sensIndex = SENS_LOW;
        } else {
            sensIndex = SENS_NORM;
        }
        calcScaleRatio();
        updateEvt = true;
    }

    // 自動モードボタン
    else if (checkBtnAction(BUTTON_AUTO, gpio, events)) {
        autoSet = !autoSet;
        updateEvt = true;
    }

    // カウントモードボタン
    else if (checkBtnAction(BUTTON_COUNT, gpio, events)) {
        if (!counting) {
            pause = false;
            counting = true;
        } else {
            counting = false;
        }
        updateEvt = true;
    }

    // クリアボタン
    else if (checkBtnAction(BUTTON_CLEAR, gpio, events)) {
        if (!gpio_get(BUTTON_POWER)) {
            autoSetWaitUs = 0;
        } else {
            undoCounter(1);
            taCounter = 0.0f;
            counterList[counterListIndex] = taCounter;
            pause = false;
        }
        updateEvt = true;
    }

    // 戻る
    else if (checkBtnAction(BUTTON_BACK, gpio, events)) {
        if (!gpio_get(BUTTON_POWER)) {
            autoSetWaitUs -= AUTO_SET_ADJ_STEP;
            if (autoSetWaitUs < 0) {
                autoSetWaitUs = 0;
            }
        } else {
            undoCounter(-1);
        }
        updateEvt = true;
    }

    // 進む
    else if (checkBtnAction(BUTTON_FORWARD, gpio, events)) {
        if (!gpio_get(BUTTON_POWER)) {
            autoSetWaitUs += AUTO_SET_ADJ_STEP;
            if (autoSetWaitUs > AUTO_SET_WAIT_MAX_uS) {
                autoSetWaitUs = AUTO_SET_WAIT_MAX_uS;
            }
        } else {
            undoCounter(1);
        }
        updateEvt = true;
    }
}

/**
 * @brief GPIO初期化
 */
static void initIO() {
    // LED
    gpio_init(LED_PIN);
    gpio_set_dir(LED_PIN, GPIO_OUT);

    // USB接続チェック用
    gpio_init(VBUS_SENSE_PIN);
    gpio_set_dir(VBUS_SENSE_PIN, GPIO_IN);

    // 電源チェック用
    adc_init();
    adc_gpio_init(VSYS_ADC_PIN);

    // GPIO
    uint size = sizeof(GPIO_PIN) / sizeof(uint);
    for (int i = 0; i < size; ++i) {
        gpio_init(GPIO_PIN[i]);
        gpio_set_dir(GPIO_PIN[i], GPIO_IN);
        gpio_pull_up(GPIO_PIN[i]);
        gpio_set_irq_enabled_with_callback(GPIO_PIN[i], GPIO_IRQ_EDGE_FALL, true, irqHandler);
        gpio_set_irq_enabled_with_callback(GPIO_PIN[i], GPIO_IRQ_EDGE_RISE, true, irqHandler);
    }
}

/**
 * @brief LCDを更新する。
 */
static void update() {
    char       *statusLabel;
    static char message[30];
    static char sensLabel[4];

    if (!powerView && gpio_get(BUTTON_POWER)) {
        if (setEvt) {
            statusLabel = "SET";
        } else {
            char tlabel[5];
            tlabel[0] = (sensIndex == SENS_LOW) ? '*' : ' ';
            tlabel[1] = (autoSet) ? 'a' : ' ';
            tlabel[2] = (counting) ? 'c' : ' ';
            tlabel[3] = (counting && pause) ? '!' : ' ';
            tlabel[4] = '\0';
            statusLabel = tlabel;
        }
        if (sens[sensIndex] < 1.0f) {
            sprintf(sensLabel, ".%02d", (int16_t)(sens[sensIndex] * 100));
        } else if (sens[sensIndex] < 10.0f) {
            sprintf(sensLabel, "%3.1f", sens[sensIndex]);
        } else {
            sprintf(sensLabel, "%3d", (int16_t)(sens[sensIndex]));
        }
        sprintf(message, "%4.2f %+6.3f %s", pcTA, pcTA - setTA, statusLabel);
        lcdPrint(1, message);
        sprintf(message, "%4.2f %s %7.3f", setTA, sensLabel, taCounter);
        lcdPrint(2, message);
    } else {
        // 電源電圧計測
        adc_select_input(VSYS_ADC_NO);
        float vsys = adc_read() * (3.3f * 3.0f / 4096.0f);
        if (usbConnected) {    // USBによる給電
            statusLabel = "USB :";
        } else {
            statusLabel = "BATT:";
            vsys += 0.18;    // ショットキーバリアダイオード分
        }
        sprintf(message, "%s%4.2fv      ", statusLabel, vsys);
        lcdPrint(1, message);
        sprintf(message, "auto set:%3.1fsec ", (float)(autoSetWaitUs / 1000.0f / 1000.0f));
        lcdPrint(2, message);
    }
}

/**
 * @brief 状態を確認する。
 */
static void checkState() {
    static int64_t lastUs = 0;
    if (autoSet && !standby) {
        int64_t nowUs = time_us_64();
        if ((taRange > TA_PAUSE_RANGE)) {    // カウントを停止
            pause = true;
        }

        // 自動セット
        if (outOfRange) {
            if (nowUs >= (lastUs + autoSetWaitUs)) {
                setEvt = true;
                lastUs = nowUs;
            }
        } else {
            lastUs = nowUs;
        }
    }
}

/**
 * @brief SAMPLING_uS間 メーターを駆動する。
 */
static void driveNeedle() {
    samplingCount = 0;
    timeExceeded = false;
    outOfRange = true;

    if (standby) {
        adcStart();
        standby = false;
    }

    float scaleSum = 0.0f;
    float pcTASum = 0.0f;
    add_alarm_in_us(SAMPLING_uS, alarmHandler, NULL, false);
    while (!timeExceeded) {
        ++samplingCount;
        pcV = adcRead();
        pcTA = v2ta(pcV);
        pcTASum += pcTA;
        float scale = (setTA - pcTA) * taScaleRatio;    // セット位置からの相対位置
        scale += METER_SCALE_SET;
        scaleSum += scale;
        if (pcTA < PC_TA_MAX) {
            if (scale > 0.0f) {
                if (scale <= (METER_SCALE_SIZE + 1.5f)) outOfRange = false;        //+1.5:電流計の誤差を考慮
                dacPos = scale * ((DAC_MAX_VALUE) / (METER_SCALE_SIZE + 3.0f));    //+3:右端を超えて振れるようにする
            } else {
                dacPos = 0.0f;
            }
            dacPos = meterDacSet(dacPos);
            sleep_us(ADC_WAIT_US);
        } else {
            adcStop();    // 節電
            standby = true;
            pause = true;
            dacPos = meterDacSet(0.0f);
            sleep_us(SAMPLING_uS * 2);
            break;
        }
    }

    // SAMPLING_uSの平均値を計算
    pcTA = pcTASum / samplingCount;
    meterValue = scaleSum / samplingCount;
}

/**
 * @brief メイン
 */
int main() {
    sleep_ms(100);
    initIO();
    initDevice();

    sleep_ms(100);
    autoSet = false;
    setEvt = false;
    sensEvt = false;
    updateEvt = true;
    powerView = true;
    sens[SENS_NORM] = SENS_NORM_INITIAL;
    sens[SENS_LOW] = SENS_LOW_INITIAL;
    usbConnected = gpio_get(VBUS_SENSE_PIN);

    bool debugout = (usbConnected && !gpio_get(BUTTON_POWER));
    if (debugout) {
        stdio_init_all();
    }

    // 設定モードを表示
    update();
    uint64_t waitUs = time_us_64() + POWER_VIEW_WAIT_uS;
    while (powerView) {
        if ((time_us_64() > waitUs) || !gpio_get(BUTTON_POWER)) {
            powerView = false;
            break;
        } else if (!gpio_get(BUTTON_SET) && usbConnected) {
            lcdPrint(1, "Device Firmware ");
            lcdPrint(2, "Update          ");
            reset_usb_boot(0, 0);
        }
        sleep_ms(1);
    }

    uint64_t netxUpdateUs = 0;
    meterValue = METER_SCALE_SET;
    calcScaleRatio();

    while (true) {
        checkState();
        if (debugout) {
            char   *label = " ";
            int32_t sps = samplingCount * (1000000 / SAMPLING_uS);
            if (setEvt) label = "SET";
            printf("SPS:%-03d pcV:%-8.3f pcTA:%-06.4f setTA:%-06.4f sense:%-05.2f sensGain:%-07.3f taDiv:%-07.4f meterValue:%-07.1f dacPos:%04d %s\n",
                   sps, pcV * 1000.0f, pcTA, setTA, sens[sensIndex], sensGain, pcTA - setTA, meterValue, (int16_t)dacPos, label);
        }
        // 感度変更イベント
        if (sensEvt) {
            calcScaleRatio();
            update();
            netxUpdateUs = time_us_64() + UPDATE_uS;
            sensEvt = false;
        }
        // セットイベント
        else if (setEvt) {
            gpio_put(LED_PIN, true);
            update();
            countTA();
            setTA = pcTA;
            calcScaleRatio();
            clearTaHist();
            netxUpdateUs = time_us_64() + UPDATE_uS;    // UPDATE_uSの間 "SET"を表示する
            outOfRange = false;
            setEvt = false;
            gpio_put(LED_PIN, false);
        }
        // 表示更新イベント
        else if (updateEvt) {
            update();
            netxUpdateUs = time_us_64() + UPDATE_uS;
            updateEvt = false;
        }
        // イベント外
        else {
            driveNeedle();
            addTaList(pcTA);

            // UPDATE_uS以上経過していたらLCDを更新
            uint64_t nowUs = time_us_64();
            if (nowUs >= netxUpdateUs) {
                update();
                netxUpdateUs = nowUs + UPDATE_uS;
            }
        }
    }
    return 0;
}